const inviteHeaders = "MIME-Version: 1.0\r\nContent-Type: text/x-msmsgsinvite; charset=UTF-8\r\n\r\n";

function MSNSession(client, email) {
  // object-specific
  this.client = client;

  // user-related
  this.contacts = new Array();
  this.contacts.push(this.client.getContact(email));

  // connection-related
  this.trID = 0;

  // session-related
  this.sessionID = 0;
  this.sessionStatus = 0;
  this.dontSend = false;
  this.msgBuffer = new Array();
}

MSNSession.prototype.__proto__ = BaseConnection;

MSNSession.prototype.onProtocolMessageReceived = function(msg, payload) {
  switch (msg[0]) {
    case "USR":
      if (msg[2] == "OK") {
        this.invite(this.contacts[0].email);
      }
      break;
    case "CAL":
      if (msg[2] == "RINGING") {
        this.sessionID = msg[3];
      }
      break;
    case "JOI":
      var contact = this.client.getContact(msg[1]);
      if (contact == null) {
        contact = new MSNContact(msg[1], decodeURIComponent(uniConvertTo(msg[2])), "NLN", 0);
      }
      if (contact.email != this.contacts[0].email) {
        this.contacts.push(contact);
      }
      if (this.contacts.length > 1) {
        showNotificationMessage("joinMessage", contact.nickname);
      }
      this.sessionStatus = 1;
      contactsChanged();
      if (this.msgBuffer.length != 0) {
        for (var i in this.msgBuffer) {
          this.sendNMessage(this.msgBuffer[i]);
        }
        delete this.msgBuffer;
        this.msgBuffer = new Array();
      }
      break;
    case "ANS":
      this.sessionStatus = 1;
      contactsChanged();
      break;
    case "IRO":
      var contact = this.client.getContact(msg[4]);
      if (contact == null) {
        contact = new MSNContact(msg[4], decodeURIComponent(uniConvertTo(msg[5])), "NLN", 0);
      }
      if (contact.email != this.contacts[0].email) {
        this.contacts.push(contact);
      }
      break;
    case "BYE":
      var contact = this.getContact(msg[1]);
      if (this.contacts.length == 1) {
        this.sessionStatus = 2;
      }
      else {
        this.contacts.splice(this.getContactIndex(msg[1]), 1);
        showNotificationMessage("leaveMessage", contact.nickname);
        contactsChanged();
      }
      break;
    case "MSG":
      var msgHeadersEnd = payload.indexOf('\r\n\r\n');
      var msgHeaders = parseHeaders(payload.substr(0, msgHeadersEnd));
      var msgBody = payload.substr(msgHeadersEnd + 4, payload.length - msgHeadersEnd - 1);
      this.messageReceived(msg[1], msgHeaders, msgBody);
      break;
    case "NAK":
      showNotificationMessage("messageNotSent", '');
      break;
    case "217":
      showNotificationMessage("contactOffline", '');
      break;
    default:
      promptService.alert(window, "Error", "Unexpected command: " + msg[0]);
      break;
  }
}

MSNSession.prototype.answer = function(host, port, cki, sessionID) {
  this.connect(host, port);
  this.sendMsg(['ANS', ++this.trID, this.client.passport, cki, sessionID]);
  this.sessionID = sessionID;
}

MSNSession.prototype.close = function() {
  if (this.sendTimeout) {
    clearTimeout(this.sendTimeout);
  }
  if (this.isConnected == true) {
    this.sendData('OUT\r\n');
    this.disconnect();
  }
}

MSNSession.prototype.contactChanged = function(email, nickname, status) {
  var contact = this.getContact(email);
  if (contact != null) {
    contact.nickname = nickname;
    contact.status = status;
    contactsChanged();
  }
}

MSNSession.prototype.getContact = function(email) {
  var contact = null;
  for (var i = 0; i < this.contacts.length; i++) {
    if (this.contacts[i].email == email) {
      contact = this.contacts[i];
    }
  }
  return contact;
}

MSNSession.prototype.getContactCount = function() {
  return this.contacts.length;
}

MSNSession.prototype.getContactIndex = function(email) {
  var index;
  for (var i in this.contacts) {
    if (this.contacts[i].email == email) {
      index = i;
    }
  }
  return index;
}

MSNSession.prototype.getContacts = function() {
  var contacts = new Array();
  for (var i = 0; i < this.contacts.length; i++) {
    contact = this.contacts[i];
    contacts.push(new MSNContact(contact.email, contact.nickname, contact.status, contact.lists));
  }
  return contacts;
}

MSNSession.prototype.getFirstContactID = function() {
  return this.contacts[0].email;
}

MSNSession.prototype.getTransferByInvCookie = function(invCookie) {
  var transfer = null;
  var winEnum = winMediator.getEnumerator("msnmsgr:file-transfer");
  while (winEnum.hasMoreElements() == true) {
    var win = winEnum.getNext();
    if (win.getInvCookie() == invCookie) {
      transfer = win;
    }
  }
  return transfer;
}

MSNSession.prototype.handleTyping = function() {
  if (this.sessionStatus > 1 && this.contacts.length == 1) {
    this.restoreSession();
  }
  if (this.sessionStatus == 1 && this.dontSend == false) {
    this.dontSend = true;
    this.sendTyping();
    var self = this;
    this.sendTimeout = setTimeout(function() { self.dontSend = false; }, 5000);
  }
}

MSNSession.prototype.initTransfer = function(file, behindRouter) {
  var invCookie = Math.round(Math.random() * 1000000);
  window.openDialog("chrome://msnmsgr/content/filetransfer.xul", "_blank", "chrome",
    {invCookie: invCookie, file: file, send: true});
  var payload = inviteHeaders;
  payload += "Application-Name: File Transfer\r\n";
  payload += "Application-GUID: {5D3E02AB-6190-11d3-BBBB-00C04F795683}\r\n";
  payload += "Invitation-Command: INVITE\r\n";
  payload += "Invitation-Cookie: " + invCookie + '\r\n';
  payload += "Application-File: " + file.leafName + '\r\n';
  payload += "Application-FileSize: " + file.fileSize + '\r\n';
  if (behindRouter == true) {
    payload += "Connectivity: N\r\n";
  }
  this.sendNMessage(payload);
  showNotificationMessage("transferInit", file.leafName);
}

MSNSession.prototype.invite = function(email) {
  this.sendData('CAL ' + ++this.trID + ' ' + email + '\r\n');
}

MSNSession.prototype.onConnectionReset = function() {
  if (this.contacts.length == 1) {
    this.sessionStatus = 3;
  }
  else {
    this.sessionStatus = -1;
    sessionClosed();
  }
}

MSNSession.prototype.open = function(host, port, cki) {
  this.connect(host, port);
  this.sendMsg(['USR', ++this.trID, this.client.passport, cki]);
}

MSNSession.prototype.messageReceived = function(from, headers, body) {
  switch (headers["content-type"]) {
    case "text/plain; charset=UTF-8":
      var contact = this.getContact(from);
      showIM(contact.nickname, from, uniConvertTo(body), headers["x-mms-im-format"], false);
      contact.timeout = 0;
      this.updateTypingContacts();
      window.getAttention();
      break;
    case "text/x-msmsgscontrol":
      var typinguser = headers["typinguser"].toLowerCase();
      var contact = this.getContact(typinguser);
      contact.timeout = new Date().getTime();
      var self = this;
      setTimeout(function() { self.updateTypingContacts(); }, 6000);
      this.updateTypingContacts();
      break;
    case "text/x-msmsgsinvite; charset=UTF-8":
      var fields = body.split('\r\n');
      var params = parseFields(fields, ": ");
      var invCookie = params["invitation-cookie"];
      switch (params["invitation-command"]) {
        case "INVITE":
          if (params["application-guid"] == "{5D3E02AB-6190-11d3-BBBB-00C04F795683}") {
            this.transferInviteReceived(invCookie, params["application-file"], params["application-filesize"], params["connectivity"]);
          }
          else {
            this.sendCancelMessage(invCookie, "REJECT_NOT_INSTALLED");
            showNotificationMessage("inviteUnsupported", '');
          }
          break;
        case "ACCEPT":
          var transfer = this.getTransferByInvCookie(invCookie);
          if (transfer != null) {
            this.transferAcceptReceived(transfer, params["authcookie"], params["ip-address"], params["sender-connect"]);
          }
          break;
        case "CANCEL":
          var transfer = this.getTransferByInvCookie(invCookie);
          if (transfer != null) {
            transfer.cancelTransfer(false);
          }
          break;
      }
      break;
  }
}

MSNSession.prototype.reset = function() {
  this.sessionID = 0;
  this.sessionStatus = 0;
  this.trID = 0;
  this.dontSend = false;
}

MSNSession.prototype.restoreSession = function() {
  var status = this.sessionStatus;
  this.sessionStatus = 0;
  if (status == 2) {
    this.invite(this.contacts[0].email);
  }
  if (status == 3) {
    createNewSession();
  }
}

MSNSession.prototype.sendAcceptMessage = function(invCookie, authCookie, senderConnect) {
  var payload = inviteHeaders;
  if (authCookie != 0) {
    payload += "IP-Address: " + this.client.getIPAddress() + '\r\n';
    payload += "Port: 6891\r\n";
    payload += "PortX: 11178\r\n";
    payload += "AuthCookie: " + authCookie + '\r\n';
  }
  if (senderConnect == true) {
    payload += "Sender-Connect: TRUE\r\n";
  }
  payload += "Invitation-Command: ACCEPT\r\n";
  payload += "Invitation-Cookie: " + invCookie + '\r\n';
  payload += "Launch-Application: FALSE\r\n";
  payload += "Request-Data: IP-Address:\r\n";
  this.sendNMessage(payload);
}

MSNSession.prototype.sendCancelMessage = function(invCookie, cancelCode) {
  var payload = inviteHeaders;
  payload += "Invitation-Command: CANCEL\r\n";
  payload += "Invitation-Cookie: " + invCookie + "\r\n";
  payload += "Cancel-Code: " + cancelCode + "\r\n";
  this.sendNMessage(payload);
}

MSNSession.prototype.sendIM = function(msgText, mms) {
  var payload = "MIME-Version: 1.0\r\n";
  payload += "Content-Type: text/plain; charset=UTF-8\r\n";
  payload += "X-MMS-IM-Format: " + mms + "\r\n";
  payload += '\r\n';
  payload += msgText;
  this.sendNMessage(payload);
}

MSNSession.prototype.sendNMessage = function(payload) {
  if (this.sessionStatus == 1) {
    this.sendData("MSG " + ++this.trID + " N " + payload.length + "\r\n" + payload);
  }
  else {
    this.msgBuffer.push(payload);
    if (this.sessionStatus > 0 && this.contacts.length == 1) {
      this.restoreSession();
    }
  }
}

MSNSession.prototype.sendTyping = function() {
  var payload = "MIME-Version: 1.0\r\n";
  payload += "Content-Type: text/x-msmsgscontrol\r\n";
  payload += "TypingUser: " + this.client.passport + "\r\n";
  payload += '\r\n';
  var msg = "MSG " + ++this.trID + " U " + payload.length + "\r\n" + payload;
  this.sendData(msg);
}

MSNSession.prototype.transferAcceptReceived = function(transfer, authCookie, ip, senderConnect) {
  if (transfer.getSend() == true && senderConnect == null) {
    // We need to send a message with the offer to serve, then listen
    authCookie = Math.round(Math.random() * 1000000);
    this.sendAcceptMessage(transfer.getInvCookie(), authCookie, false);
  }
  transfer.connect(authCookie, ip);
}

MSNSession.prototype.transferDone = function(result, invCookie, fileName) {
  var resultLabel = '';
  switch (result) {
    case 0:
      resultLabel = "transferComplete";
      break;
    case 1:
      resultLabel = "transferCancelled";
      break;
    case 2:
      resultLabel = "transferProblem";
      this.sendCancelMessage(invCookie, "FTTIMEOUT");
      break;
    case 3:
      resultLabel = "transferCancelled";
      this.sendCancelMessage(invCookie, "TIMEOUT");
  }
  showNotificationMessage(resultLabel, fileName);
}

MSNSession.prototype.transferInviteReceived = function(invCookie, fileName, fileSize, connectivity) {
  var file = receiveFile(fileName, Math.ceil(fileSize / 1024));
  if (file != null) {
    if (connectivity == null) {
      this.sendAcceptMessage(invCookie, 0, false);
      window.openDialog("chrome://msnmsgr/content/filetransfer.xul", "_blank", "chrome",
        {invCookie: invCookie, file: file, send: false});
    }
    else {
      var authCookie = Math.round(Math.random() * 1000000);
      this.sendAcceptMessage(invCookie, authCookie, true);
      window.openDialog("chrome://msnmsgr/content/filetransfer.xul", "_blank", "chrome",
        {invCookie: invCookie, file: file, send: false}, {authCookie: authCookie});
    }
  }
  else {
    this.sendCancelMessage(invCookie, "REJECT");
  }
}

MSNSession.prototype.updateTypingContacts = function() {
  var contacts = '';
  var count = 0;
  var time = new Date().getTime();
  for (var i = 0; i < this.contacts.length; i++) {
    var contact = this.contacts[i];
    if (time <= contact.timeout + 5000) {
      count++;
      contacts += contact.nickname + ", ";
    }
  }
  if (count != 0) {
    contacts = contacts.substr(0, contacts.length - 2);
  }
  showTypingContacts(contacts, count);
}